from flama import exceptions, websockets
from flama.codecs.base import WebsocketsCodec

__all__ = ["TextCodec"]


class TextCodec(WebsocketsCodec):
    encoding = "text"

    async def decode(self, message: websockets.Message, **options):
        if "text" not in message:
            raise exceptions.DecodeError("Expected text websocket messages")

        return message["text"]



import typing

from flama import exceptions
from flama.codecs.base import HTTPCodec, WebsocketsCodec
from flama.codecs.websockets import BytesCodec

__all__ = ["ContentTypeNegotiator", "WebSocketEncodingNegotiator"]


class ContentTypeNegotiator:
    def __init__(self, codecs: typing.Optional[typing.List[HTTPCodec]] = None):
        self.codecs = codecs or []

    def negotiate(self, content_type: str = None) -> WebsocketsCodec:
        """
        Given the value of a 'Content-Type' header, return the appropriate codec for decoding the request content.
        """
        if content_type is None:
            return self.codecs[0]

        content_type = content_type.split(";")[0].strip().lower()
        main_type = content_type.split("/")[0] + "/*"
        wildcard_type = "*/*"

        for codec in self.codecs:
            if codec.media_type in (content_type, main_type, wildcard_type):
                return codec

        raise exceptions.NoCodecAvailable(f"Unsupported media in Content-Type header '{content_type}'")


class WebSocketEncodingNegotiator:
    def __init__(self, codecs: typing.Optional[typing.List[WebsocketsCodec]] = None):
        self.codecs = codecs or [BytesCodec()]

    def negotiate(self, encoding: str = None) -> WebsocketsCodec:
        """
        Given a websocket encoding, return the appropriate codec for decoding the request content.
        """
        if encoding is None:
            return self.codecs[0]

        for codec in self.codecs:
            if codec.encoding == encoding:
                return codec

        raise exceptions.NoCodecAvailable(f"Unsupported websocket encoding '{encoding}'")



import inspect
from abc import ABCMeta, abstractmethod

from flama import exceptions


class Component(metaclass=ABCMeta):
    def identity(self, parameter: inspect.Parameter) -> str:
        """
        Each component needs a unique identifier string that we use for lookups from the `state` dictionary when we run
        the dependency injection.

        :param parameter: The parameter to check if that component can handle it.
        :return: Unique identifier.
        """
        parameter_name = parameter.name.lower()
        try:
            annotation_name = parameter.annotation.__name__.lower()
        except AttributeError:
            annotation_name = parameter.annotation.__args__[0].__name__.lower()

        # If `resolve_parameter` includes `Parameter` then we use an identifier that is additionally parameterized by
        # the parameter name.
        args = inspect.signature(self.resolve).parameters.values()
        if inspect.Parameter in [arg.annotation for arg in args]:
            return annotation_name + ":" + parameter_name

        # Standard case is to use the class name, lowercased.
        return annotation_name

    def can_handle_parameter(self, parameter: inspect.Parameter) -> bool:
        """
        The default behavior is for components to handle whatever class is used as the return annotation by the
        `resolve` method.

        You can override this for more customized styles, for example if you wanted name-based parameter resolution, or
        if you want to provide a value for a range of different types.

        Eg. Include the `Request` instance for any parameter named `request`.

        :param parameter: The parameter to check if that component can handle it.
        :return: True if this component can handle the given parameter.
        """
        return_annotation = inspect.signature(self.resolve).return_annotation
        if return_annotation is inspect.Signature.empty:
            msg = (
                f'Component "{self.__class__.__name__}" must include a return annotation on the `resolve()` method, or '
                f"override `can_handle_parameter`"
            )
            raise exceptions.ConfigurationError(msg)
        return parameter.annotation is return_annotation

    @abstractmethod
    def resolve(self):
        pass



import typing

import starlette.exceptions


class DecodeError(Exception):
    """
    Raised by a Codec when `decode` fails due to malformed syntax.
    """

    def __init__(self, message, marker=None, base_format=None):
        Exception.__init__(self, message)
        self.message = message
        self.marker = marker
        self.base_format = base_format


class NoReverseMatch(Exception):
    """
    Raised by a Router when `reverse_url` is passed an invalid handler name.
    """

    ...


class NoCodecAvailable(Exception):
    ...


class ConfigurationError(Exception):
    ...


class ComponentNotFound(ConfigurationError):
    def __init__(self, parameter: str, resolver: typing.Optional[str] = None, *args, **kwargs):
        self.parameter = parameter
        self.resolver = resolver
        super().__init__(*args, **kwargs)

    def __str__(self):
        msg = f'No component able to handle parameter "{self.parameter}"'
        if self.resolver:
            msg += f' in function "{self.resolver}"'

        return msg


class WebSocketException(Exception):
    def __init__(self, close_code: int):
        self.close_code = close_code


class WebSocketConnectionException(Exception):
    ...


class HTTPException(starlette.exceptions.HTTPException):
    ...


class ValidationError(HTTPException):
    def __init__(self, detail: typing.Union[str, typing.Dict[str, typing.List[str]]], status_code: int = 400):
        super().__init__(status_code=status_code, detail=detail)


class InputValidationError(ValidationError):
    ...


class OutputValidationError(ValidationError):
    ...



import inspect
import itertools
import os
import typing
from collections import defaultdict
from string import Template

import marshmallow
from starlette import routing, schemas
from starlette.responses import HTMLResponse

from flama.responses import APIError
from flama.types import EndpointInfo
from flama.utils import dict_safe_add

try:
    import apispec
except Exception:  # pragma: no cover
    apispec = None  # type: ignore

try:
    import yaml
except Exception:  # pragma: no cover
    yaml = None  # type: ignore

__all__ = ["OpenAPIResponse", "SchemaGenerator", "SchemaMixin"]


if yaml is not None and apispec is not None:
    from apispec.yaml_utils import YAMLDumper as BaseYAMLDumper

    class YAMLDumper(BaseYAMLDumper):
        def ignore_aliases(self, data):
            return True


class OpenAPIResponse(schemas.OpenAPIResponse):
    def render(self, content: typing.Any) -> bytes:
        assert yaml is not None, "`pyyaml` must be installed to use OpenAPIResponse."
        assert apispec is not None, "`apispec` must be installed to use OpenAPIResponse."
        assert isinstance(content, dict), "The schema passed to OpenAPIResponse should be a dictionary."

        return yaml.dump(content, default_flow_style=False, Dumper=YAMLDumper).encode("utf-8")


class SchemaRegistry(dict):
    def __init__(self, spec, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spec = spec
        self.openapi = self.spec.plugins[0].openapi

    def __getitem__(self, item):
        is_class = inspect.isclass(item)
        schema_class = item if is_class else item.__class__

        try:
            schema = super().__getitem__(schema_class)
        except KeyError:
            self.spec.components.schema(name=schema_class.__name__, schema=schema_class)
            schema = self.openapi.resolve_schema_dict(schema_class)
            super().__setitem__(schema_class, schema)

        if not is_class:
            schema = self.openapi.resolve_schema_dict(item)

        return schema


class SchemaGenerator(schemas.BaseSchemaGenerator):
    def __init__(self, title: str, version: str, description: str, openapi_version="3.0.0"):
        assert apispec is not None, "`apispec` must be installed to use SchemaGenerator."

        from apispec.ext.marshmallow import MarshmallowPlugin

        self.spec = apispec.APISpec(
            title=title,
            version=version,
            openapi_version=openapi_version,
            info={"description": description},
            plugins=[MarshmallowPlugin()],
        )
        self.openapi = self.spec.plugins[0].openapi

        # Builtin definitions
        self.schemas = SchemaRegistry(self.spec)

    def get_endpoints(
        self, routes: typing.List[routing.BaseRoute], base_path: str = ""
    ) -> typing.Dict[str, typing.Sequence[EndpointInfo]]:
        """
        Given the routes, yields the following information:

        - path
            eg: /users/
        - http_method
            one of 'get', 'post', 'put', 'patch', 'delete', 'options'
        - func
            method ready to extract the docstring
        """
        endpoints_info: typing.Dict[str, typing.Sequence[EndpointInfo]] = defaultdict(list)

        for route in routes:
            if isinstance(route, routing.Route) and route.include_in_schema:
                _, path, _ = routing.compile_path(base_path + route.path)

                if inspect.isfunction(route.endpoint) or inspect.ismethod(route.endpoint):
                    for method in route.methods or ["GET"]:
                        if method == "HEAD":
                            continue

                        endpoints_info[path].append(
                            EndpointInfo(
                                path=path,
                                method=method.lower(),
                                func=route.endpoint,
                                query_fields=route.query_fields.get(method, {}),
                                path_fields=route.path_fields.get(method, {}),
                                body_field=route.body_field.get(method),
                                output_field=route.output_field.get(method),
                            )
                        )
                else:
                    for method in ["get", "post", "put", "patch", "delete", "options"]:
                        if not hasattr(route.endpoint, method):
                            continue

                        func = getattr(route.endpoint, method)
                        endpoints_info[path].append(
                            EndpointInfo(
                                path=path,
                                method=method.lower(),
                                func=func,
                                query_fields=route.query_fields.get(method.upper(), {}),
                                path_fields=route.path_fields.get(method.upper(), {}),
                                body_field=route.body_field.get(method.upper()),
                                output_field=route.output_field.get(method.upper()),
                            )
                        )
            elif isinstance(route, routing.Mount):
                endpoints_info.update(self.get_endpoints(route.routes, base_path=route.path))

        return endpoints_info

    def _add_endpoint_parameters(self, endpoint: EndpointInfo, schema: typing.Dict):
        schema["parameters"] = [
            self.openapi.field2parameter(field.schema, name=field.name, default_in=field.location.name)
            for field in itertools.chain(endpoint.query_fields.values(), endpoint.path_fields.values())
        ]

    def _add_endpoint_body(self, endpoint: EndpointInfo, schema: typing.Dict):
        dict_safe_add(
            schema, self.schemas[endpoint.body_field.schema], "requestBody", "content", "application/json", "schema"
        )

    def _add_endpoint_response(self, endpoint: EndpointInfo, schema: typing.Dict):
        response_codes = list(schema.get("responses", {}).keys())
        main_response = response_codes[0] if response_codes else 200

        dict_safe_add(
            schema,
            self.schemas[endpoint.output_field],
            "responses",
            main_response,
            "content",
            "application/json",
            "schema",
        )

    def _add_endpoint_default_response(self, schema: typing.Dict):
        dict_safe_add(schema, self.schemas[APIError], "responses", "default", "content", "application/json", "schema")

        # Default description
        schema["responses"]["default"]["description"] = schema["responses"]["default"].get(
            "description", "Unexpected error."
        )

    def get_endpoint_schema(self, endpoint: EndpointInfo) -> typing.Dict[str, typing.Any]:
        schema = self.parse_docstring(endpoint.func)

        # Query and Path parameters
        if endpoint.query_fields or endpoint.path_fields:
            self._add_endpoint_parameters(endpoint, schema)

        # Body
        if endpoint.body_field:
            self._add_endpoint_body(endpoint, schema)

        # Response
        if endpoint.output_field and (
            (inspect.isclass(endpoint.output_field) and issubclass(endpoint.output_field, marshmallow.Schema))
            or isinstance(endpoint.output_field, marshmallow.Schema)
        ):
            self._add_endpoint_response(endpoint, schema)

        # Default response
        self._add_endpoint_default_response(schema)

        return schema

    def get_schema(self, routes: typing.List[routing.BaseRoute]) -> typing.Dict[str, typing.Any]:
        endpoints_info = self.get_endpoints(routes)

        for path, endpoints in endpoints_info.items():
            self.spec.path(path=path, operations={e.method: self.get_endpoint_schema(e) for e in endpoints})

        return self.spec.to_dict()


class SchemaMixin:
    def add_schema_docs_routes(
        self,
        title: str = "",
        version: str = "",
        description: str = "",
        schema: typing.Optional[str] = "/schema/",
        docs: typing.Optional[str] = "/docs/",
        redoc: typing.Optional[str] = None,
    ):
        # Schema
        self.title = title
        self.version = version
        self.description = description
        self.schema_url = schema
        if self.schema_url:
            self.add_schema_route()

        # Docs (Swagger UI)
        self.docs_url = docs
        if self.docs_url:
            self.add_docs_route()

        # Redoc
        self.redoc_url = redoc
        if self.redoc_url:
            self.add_redoc_route()

    @property
    def schema_generator(self):
        if not hasattr(self, "_schema_generator"):
            self._schema_generator = SchemaGenerator(
                title=self.title, version=self.version, description=self.description
            )

        return self._schema_generator

    @property
    def schema(self):
        return self.schema_generator.get_schema(self.routes)

    def add_schema_route(self):
        def schema():
            return OpenAPIResponse(self.schema)

        self.add_route(path=self.schema_url, route=schema, methods=["GET"], include_in_schema=False)

    def add_docs_route(self):
        def swagger_ui() -> HTMLResponse:
            with open(os.path.join(os.path.dirname(__file__), "templates/swagger_ui.html")) as f:
                content = Template(f.read()).substitute(title=self.title, schema_url=self.schema_url)

            return HTMLResponse(content)

        self.add_route(path=self.docs_url, route=swagger_ui, methods=["GET"], include_in_schema=False)

    def add_redoc_route(self):
        def redoc() -> HTMLResponse:
            with open(os.path.join(os.path.dirname(__file__), "templates/redoc.html")) as f:
                content = Template(f.read()).substitute(title=self.title, schema_url=self.schema_url)

            return HTMLResponse(content)

        self.add_route(path=self.redoc_url, route=redoc, methods=["GET"], include_in_schema=False)



import typing

from flama import http, websockets


class Codec:
    async def decode(self, item: typing.Any, **options):
        raise NotImplementedError()

    async def encode(self, item: typing.Any, **options):
        raise NotImplementedError()


class HTTPCodec(Codec):
    media_type = None

    async def decode(self, request: http.Request, **options):
        raise NotImplementedError()

    async def encode(self, item: typing.Any, **options):
        raise NotImplementedError()


class WebsocketsCodec(Codec):
    encoding = None

    async def decode(self, item: websockets.Message, **options):
        raise NotImplementedError()

    async def encode(self, item: typing.Any, **options):
        raise NotImplementedError()



import json

from flama import exceptions, websockets
from flama.codecs.base import WebsocketsCodec

__all__ = ["JSONCodec"]


class JSONCodec(WebsocketsCodec):
    encoding = "json"

    async def decode(self, message: websockets.Message, **options):
        if message.get("text") is not None:
            text = message["text"]
        else:
            text = message["bytes"].decode("utf-8")

        try:
            return json.loads(text)
        except json.decoder.JSONDecodeError:
            raise exceptions.DecodeError("Malformed JSON data received")



import datetime
import inspect
import typing
import uuid

import marshmallow
from starlette import status

from flama import codecs, exceptions, http, websockets
from flama.components import Component
from flama.exceptions import WebSocketException
from flama.negotiation import ContentTypeNegotiator, WebSocketEncodingNegotiator
from flama.routing import Route
from flama.types import OptBool, OptDate, OptDateTime, OptFloat, OptInt, OptStr, OptUUID

ValidatedPathParams = typing.NewType("ValidatedPathParams", dict)
ValidatedQueryParams = typing.NewType("ValidatedQueryParams", dict)
ValidatedRequestData = typing.TypeVar("ValidatedRequestData")


class RequestDataComponent(Component):
    def __init__(self):
        self.negotiator = ContentTypeNegotiator(
            [codecs.JSONDataCodec(), codecs.URLEncodedCodec(), codecs.MultiPartCodec()]
        )

    def can_handle_parameter(self, parameter: inspect.Parameter):
        return parameter.annotation is http.RequestData

    async def resolve(self, request: http.Request):
        content_type = request.headers.get("Content-Type")

        try:
            codec = self.negotiator.negotiate(content_type)
        except exceptions.NoCodecAvailable:
            raise exceptions.HTTPException(415)

        try:
            return await codec.decode(request)
        except exceptions.DecodeError as exc:
            raise exceptions.HTTPException(400, detail=str(exc))


class WebSocketMessageDataComponent(Component):
    def __init__(self):
        self.negotiator = WebSocketEncodingNegotiator([codecs.BytesCodec(), codecs.TextCodec(), codecs.JSONCodec()])

    def can_handle_parameter(self, parameter: inspect.Parameter):
        return parameter.annotation is websockets.Data

    async def resolve(self, message: websockets.Message, websocket_encoding: websockets.Encoding):
        try:
            codec = self.negotiator.negotiate(websocket_encoding)
            return await codec.decode(message)
        except (exceptions.NoCodecAvailable, exceptions.DecodeError):
            raise WebSocketException(close_code=status.WS_1003_UNSUPPORTED_DATA)


class ValidatePathParamsComponent(Component):
    async def resolve(self, request: http.Request, route: Route, path_params: http.PathParams) -> ValidatedPathParams:
        validator = type(
            "Validator", (marshmallow.Schema,), {f.name: f.schema for f in route.path_fields[request.method].values()}
        )

        try:
            path_params = validator().load(path_params)
        except marshmallow.ValidationError as exc:
            raise exceptions.InputValidationError(detail=exc.normalized_messages())
        return ValidatedPathParams(path_params)


class ValidateQueryParamsComponent(Component):
    def resolve(self, request: http.Request, route: Route, query_params: http.QueryParams) -> ValidatedQueryParams:
        validator = type(
            "Validator", (marshmallow.Schema,), {f.name: f.schema for f in route.query_fields[request.method].values()}
        )

        try:
            query_params = validator().load(dict(query_params), unknown=marshmallow.EXCLUDE)
        except marshmallow.ValidationError as exc:
            raise exceptions.InputValidationError(detail=exc.normalized_messages())
        return ValidatedQueryParams(query_params)


class ValidateRequestDataComponent(Component):
    def can_handle_parameter(self, parameter: inspect.Parameter):
        return parameter.annotation is ValidatedRequestData

    def resolve(self, request: http.Request, route: Route, data: http.RequestData):
        if not route.body_field[request.method] or not route.body_field[request.method].schema:
            return data

        validator = route.body_field[request.method].schema

        try:
            return validator.load(data)
        except marshmallow.ValidationError as exc:
            raise exceptions.InputValidationError(detail=exc.normalized_messages())


class PrimitiveParamComponent(Component):
    def can_handle_parameter(self, parameter: inspect.Parameter):
        return parameter.annotation in (
            str,
            int,
            float,
            bool,
            OptStr,
            OptInt,
            OptFloat,
            OptBool,
            parameter.empty,
            http.QueryParam,
            http.PathParam,
            uuid.UUID,
            datetime.date,
            datetime.datetime,
        )

    def resolve(
        self, parameter: inspect.Parameter, path_params: ValidatedPathParams, query_params: ValidatedQueryParams
    ):
        params = path_params if (parameter.name in path_params) else query_params

        if parameter.annotation in (OptInt, OptFloat, OptBool, OptStr) or parameter.default is not parameter.empty:
            kwargs = {"missing": parameter.default if parameter.default is not parameter.empty else None}
        else:
            kwargs = {"required": True}

        param_validator = {
            inspect.Signature.empty: marshmallow.fields.Field,
            int: marshmallow.fields.Integer,
            float: marshmallow.fields.Number,
            bool: marshmallow.fields.Boolean,
            str: marshmallow.fields.String,
            uuid.UUID: marshmallow.fields.UUID,
            datetime.date: marshmallow.fields.Date,
            datetime.datetime: marshmallow.fields.DateTime,
            OptInt: marshmallow.fields.Integer,
            OptFloat: marshmallow.fields.Number,
            OptBool: marshmallow.fields.Boolean,
            OptStr: marshmallow.fields.String,
            http.QueryParam: marshmallow.fields.String,
            http.PathParam: marshmallow.fields.String,
            OptUUID: marshmallow.fields.UUID,
            OptDate: marshmallow.fields.Date,
            OptDateTime: marshmallow.fields.DateTime,
        }[parameter.annotation](**kwargs)

        validator = type("Validator", (marshmallow.Schema,), {parameter.name: param_validator})

        try:
            params = validator().load(params, unknown=marshmallow.EXCLUDE)
        except marshmallow.ValidationError as exc:
            raise exceptions.InputValidationError(detail=exc.normalized_messages())
        return params.get(parameter.name, parameter.default)


class CompositeParamComponent(Component):
    def can_handle_parameter(self, parameter: inspect.Parameter):
        return inspect.isclass(parameter.annotation) and issubclass(parameter.annotation, marshmallow.Schema)

    def resolve(self, parameter: inspect.Parameter, data: ValidatedRequestData):
        return data


VALIDATION_COMPONENTS = (
    RequestDataComponent(),
    WebSocketMessageDataComponent(),
    ValidatePathParamsComponent(),
    ValidateQueryParamsComponent(),
    ValidateRequestDataComponent(),
    PrimitiveParamComponent(),
    CompositeParamComponent(),
)



import typing

import marshmallow

from flama.responses import APIResponse

__all__ = ["LimitOffsetSchema", "LimitOffsetResponse"]


class LimitOffsetMeta(marshmallow.Schema):
    limit = marshmallow.fields.Integer(title="limit", description="Number of retrieved items")
    offset = marshmallow.fields.Integer(title="offset", description="Collection offset")
    count = marshmallow.fields.Integer(title="count", description="Total number of items", allow_none=True)


class LimitOffsetSchema(marshmallow.Schema):
    meta = marshmallow.fields.Nested(LimitOffsetMeta)
    data = marshmallow.fields.List(marshmallow.fields.Dict())


class LimitOffsetResponse(APIResponse):
    """
    Response paginated based on a limit of elements and an offset.

    First 10 elements:
        /resource?offset=0&limit=10
    Elements 20-30:
        /resource?offset=20&limit=10
    """

    default_limit = 10

    def __init__(
        self,
        schema: marshmallow.Schema,
        offset: typing.Optional[typing.Union[int, str]] = None,
        limit: typing.Optional[typing.Union[int, str]] = None,
        count: typing.Optional[bool] = True,
        **kwargs
    ):
        self.offset = int(offset) if offset is not None else 0
        self.limit = int(limit) if limit is not None else self.default_limit
        self.count = count
        super().__init__(schema=schema, **kwargs)

    def render(self, content: typing.Sequence):
        init = self.offset
        end = self.offset + self.limit
        return super().render(
            {
                "meta": {"limit": self.limit, "offset": self.offset, "count": len(content) if self.count else None},
                "data": content[init:end],
            }
        )



import asyncio
import functools

import marshmallow

from flama.pagination.limit_offset import LimitOffsetResponse, LimitOffsetSchema
from flama.pagination.page_number import PageNumberResponse, PageNumberSchema
from flama.validation import get_output_schema

try:
    import forge
except Exception:  # pragma: no cover
    forge = None  # type: ignore

__all__ = ["Paginator"]


class Paginator:
    @classmethod
    def page_number(cls, func):
        """
        Decorator for adding pagination behavior to a view. That decorator produces a view based on page numbering and
        it adds three query parameters to control the pagination: page, page_size and count. Page has a default value of
        first page, page_size default value is defined in
        :class:`PageNumberResponse` and count defines if the response will define
        the total number of elements.

        The output schema is also modified by :class:`PageNumberSchema`, creating
        a new schema based on it but using the old output schema as the content of its data field.

        :param func: View to be decorated.
        :return: Decorated view.
        """
        assert forge is not None, "`python-forge` must be installed to use OpenAPIResponse."

        resource_schema = get_output_schema(func)

        schema = type(
            "PageNumberPaginated" + resource_schema.__class__.__name__,  # Add a prefix to avoid collision
            (PageNumberSchema,),
            {"data": marshmallow.fields.Nested(resource_schema, many=True)},  # Replace generic with resource schema
        )()

        forge_revision_list = (
            forge.copy(func),
            forge.insert(forge.arg("page", default=None, type=int), index=-1),
            forge.insert(forge.arg("page_size", default=None, type=int), index=-1),
            forge.insert(forge.arg("count", default=True, type=bool), index=-1),
            forge.delete("kwargs"),
            forge.returns(schema),
        )

        try:
            if asyncio.iscoroutinefunction(func):

                @forge.compose(*forge_revision_list)
                @functools.wraps(func)
                async def decorator(*args, page: int = None, page_size: int = None, count: bool = True, **kwargs):
                    return PageNumberResponse(
                        schema=schema, page=page, page_size=page_size, count=count, content=await func(*args, **kwargs)
                    )

            else:

                @forge.compose(*forge_revision_list)
                @functools.wraps(func)
                def decorator(*args, page: int = None, page_size: int = None, count: bool = True, **kwargs):
                    return PageNumberResponse(
                        schema=schema, page=page, page_size=page_size, count=count, content=func(*args, **kwargs)
                    )

        except ValueError as e:
            raise TypeError("Paginated views must define **kwargs param") from e

        return decorator

    @classmethod
    def limit_offset(cls, func):
        """
        Decorator for adding pagination behavior to a view. That decorator produces a view based on limit-offset and
        it adds three query parameters to control the pagination: limit, offset and count. Offset has a default value of
        zero to start with the first element of the collection, limit default value is defined in
        :class:`LimitOffsetResponse` and count defines if the response will
        define the total number of elements.

        The output schema is also modified by :class:`LimitOffsetSchema`,
        creating a new schema based on it but using the old output schema as the content of its data field.

        :param func: View to be decorated.
        :return: Decorated view.
        """
        assert forge is not None, "`python-forge` must be installed to use OpenAPIResponse."

        resource_schema = get_output_schema(func)

        schema = type(
            "LimitOffsetPaginated" + resource_schema.__class__.__name__,  # Add a prefix to avoid collision
            (LimitOffsetSchema,),
            {"data": marshmallow.fields.Nested(resource_schema, many=True)},  # Replace generic with resource schema
        )()

        forge_revision_list = (
            forge.copy(func),
            forge.insert(forge.arg("limit", default=None, type=int), index=-1),
            forge.insert(forge.arg("offset", default=None, type=int), index=-1),
            forge.insert(forge.arg("count", default=True, type=bool), index=-1),
            forge.delete("kwargs"),
            forge.returns(schema),
        )

        try:
            if asyncio.iscoroutinefunction(func):

                @forge.compose(*forge_revision_list)
                @functools.wraps(func)
                async def decorator(*args, limit: int = None, offset: int = None, count: bool = True, **kwargs):
                    return LimitOffsetResponse(
                        schema=schema, limit=limit, offset=offset, count=count, content=await func(*args, **kwargs)
                    )

            else:

                @forge.compose(*forge_revision_list)
                @functools.wraps(func)
                def decorator(*args, limit: int = None, offset: int = None, count: bool = True, **kwargs):
                    return LimitOffsetResponse(
                        schema=schema, limit=limit, offset=offset, count=count, content=func(*args, **kwargs)
                    )

        except ValueError as e:
            raise TypeError("Paginated views must define **kwargs param") from e

        return decorator



import asyncio

from starlette import status
from starlette.concurrency import run_in_threadpool
from starlette.endpoints import HTTPEndpoint as BaseHTTPEndpoint
from starlette.endpoints import WebSocketEndpoint as BaseWebSocketEndpoint
from starlette.requests import Request
from starlette.websockets import WebSocket, WebSocketState

from flama import exceptions, websockets
from flama.responses import APIResponse
from flama.validation import get_output_schema

__all__ = ["HTTPEndpoint", "WebSocketEndpoint"]


class HTTPEndpoint(BaseHTTPEndpoint):
    async def dispatch(self) -> None:
        request = Request(self.scope, receive=self.receive)
        app = self.scope["app"]

        route, route_scope = app.router.get_route_from_scope(self.scope)

        state = {
            "scope": self.scope,
            "receive": self.receive,
            "send": self.send,
            "exc": None,
            "app": app,
            "path_params": route_scope["path_params"],
            "route": route,
            "request": request,
        }
        handler_name = "get" if request.method == "HEAD" else request.method.lower()
        handler = getattr(self, handler_name, self.method_not_allowed)

        injected_func = await app.injector.inject(handler, state)
        if asyncio.iscoroutinefunction(handler):
            response = await injected_func()
        else:
            response = await run_in_threadpool(injected_func)

        # Wrap response data with a proper response class
        if isinstance(response, (dict, list)):
            response = APIResponse(content=response, schema=get_output_schema(handler))
        elif isinstance(response, str):
            response = APIResponse(content=response)
        elif response is None:
            response = APIResponse(content="")

        await response(self.scope, self.receive, self.send)


class WebSocketEndpoint(BaseWebSocketEndpoint):
    async def dispatch(self) -> None:
        app = self.scope["app"]
        websocket = WebSocket(self.scope, self.receive, self.send)

        route, route_scope = app.router.get_route_from_scope(self.scope)

        state = {
            "scope": self.scope,
            "receive": self.receive,
            "send": self.send,
            "exc": None,
            "app": app,
            "path_params": route_scope["path_params"],
            "route": route,
            "websocket": websocket,
            "websocket_encoding": self.encoding,
            "websocket_code": status.WS_1000_NORMAL_CLOSURE,
            "websocket_message": None,
        }

        try:
            on_connect = await app.injector.inject(self.on_connect, state)
            await on_connect()
        except Exception as e:
            raise exceptions.WebSocketConnectionException("Error connecting socket") from e

        try:
            state["websocket_message"] = await websocket.receive()

            while websocket.client_state == WebSocketState.CONNECTED:
                on_receive = await app.injector.inject(self.on_receive, state)
                await on_receive()
                state["websocket_message"] = await websocket.receive()

            state["websocket_code"] = int(state["websocket_message"].get("code", status.WS_1000_NORMAL_CLOSURE))
        except exceptions.WebSocketException as e:
            state["websocket_code"] = e.close_code
        except Exception as e:
            state["websocket_code"] = status.WS_1011_INTERNAL_ERROR
            raise e from None
        finally:
            on_disconnect = await app.injector.inject(self.on_disconnect, state)
            await on_disconnect()

    async def on_connect(self, websocket: websockets.WebSocket) -> None:
        """Override to handle an incoming websocket connection"""
        await websocket.accept()

    async def on_receive(self, websocket: websockets.WebSocket, data: websockets.Data) -> None:
        """Override to handle an incoming websocket message"""

    async def on_disconnect(self, websocket: websockets.WebSocket, websocket_code: websockets.Code) -> None:
        """Override to handle a disconnecting websocket"""
        await websocket.close(websocket_code)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['flama',
 'flama.codecs',
 'flama.codecs.http',
 'flama.codecs.websockets',
 'flama.components',
 'flama.pagination']

package_data = \
{'': ['*'], 'flama': ['templates/*']}

install_requires = \
['marshmallow>=3.0.0rc6,<4.0', 'starlette>=0.12.0,<0.13.0']

extras_require = \
{'full': ['python-forge>=18.6,<19.0',
          'apispec>=1.0.0,<2.0.0',
          'pyyaml>=3.13,<4.0',
          'sqlalchemy>=1.2,<2.0',
          'databases>=0.1.9,<0.2.0']}

setup_kwargs = {
    'name': 'flama',
    'version': '0.9.0',
    'description': 'Fire up your API with this flamethrower',
    'long_description': '<p align="center">\n  <a href="https://flama.perdy.io"><img src="https://raw.githubusercontent.com/perdy/flama/master/docs/images/logo.png" alt=\'Flama\'></a>\n</p>\n<p align="center">\n    &#128293; <em>Fire up your API with this flamethrower.</em>\n</p>\n<p align="center">\n<a href="https://circleci.com/gh/perdy/flama">\n    <img src="https://img.shields.io/circleci/project/github/perdy/flama/master.svg" alt="Build Status">\n</a>\n<a href="https://codecov.io/gh/perdy/flama">\n    <img src="https://codecov.io/gh/perdy/flama/branch/master/graph/badge.svg" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/flama/">\n    <img src="https://badge.fury.io/py/flama.svg" alt="Package version">\n</a>\n</p>\n\n---\n\n**Documentation**: [https://flama.perdy.io](https://flama.perdy.io)\n\n---\n\n# Flama\n\nFlama aims to bring a layer on top of [Starlette] to provide an **easy to learn** and **fast to develop** approach for \nbuilding **highly performant** GraphQL and REST APIs. In the same way of Starlette is, Flama is a perfect option for \ndeveloping **asynchronous** and **production-ready** services. \n\nAmong other characteristics it provides the following:\n\n* **Generic classes** for API resources that provides standard CRUD methods over SQLAlchemy tables.\n* **Schema system** based on [Marshmallow] that allows to **declare** the inputs and outputs of endpoints and provides \na reliable way of **validate** data against those schemas.\n* **Dependency Injection** that ease the process of managing parameters needed in endpoints. Flama ASGI objects \nlike `Request`, `Response`, `Session` and so on are defined as components and ready to be injected in your endpoints.\n* **Components** as the base of the plugin ecosystem, allowing you to create custom or use those already defined in \nyour endpoints, injected as parameters.\n* **Auto generated API schema** using OpenAPI standard. It uses the schema system of your endpoints to extract all the \nnecessary information to generate your API Schema.\n* **Auto generated docs** providing a [Swagger UI] or [ReDoc] endpoint.\n* **Pagination** automatically handled using multiple methods such as limit and offset, page numbers...\n\n## Requirements\n\n* [Python] 3.6+\n* [Starlette] 0.10+\n* [Marshmallow] 3.0+\n\n## Installation\n\n```console\n$ pip install flama\n```\n\n## Example\n\n```python\nfrom marshmallow import Schema, fields, validate\nfrom flama.applications import Flama\n\n\n# Data Schema\nclass Puppy(Schema):\n    id = fields.Integer()\n    name = fields.String()\n    age = fields.Integer(validate=validate.Range(min=0))\n\n\n# Database\npuppies = [\n    {"id": 1, "name": "Canna", "age": 6},\n    {"id": 2, "name": "Sandy", "age": 12},\n]\n\n\n# Application\napp = Flama(\n    components=[],      # Without custom components\n    title="Foo",        # API title\n    version="0.1",      # API version\n    description="Bar",  # API description\n    schema="/schema/",  # Path to expose OpenAPI schema\n    docs="/docs/",      # Path to expose Swagger UI docs\n    redoc="/redoc/",    # Path to expose ReDoc docs\n)\n\n\n# Views\n@app.route("/", methods=["GET"])\ndef list_puppies(name: str = None) -> Puppy(many=True):\n    """\n    description:\n        List the puppies collection. There is an optional query parameter that \n        specifies a name for filtering the collection based on it.\n    responses:\n        200:\n            description: List puppies.\n    """\n    return [puppy for puppy in puppies if puppy["name"] == name]\n    \n\n@app.route("/", methods=["POST"])\ndef create_puppy(puppy: Puppy) -> Puppy:\n    """\n    description:\n        Create a new puppy using data validated from request body and add it \n        to the collection.\n    responses:\n        200:\n            description: Puppy created successfully.\n    """\n    puppies.append(puppy)\n    \n    return puppy\n```\n\n## Dependencies\n\nFollowing Starlette philosophy Flama reduce the number of hard dependencies to those that are used as the core:\n\n* [`starlette`][Starlette] - Flama is a layer on top of it.\n* [`marshmallow`][Marshmallow] - Flama data schemas and validation.\n\nIt does not have any more hard dependencies, but some of them are necessaries to use some features:\n\n* [`pyyaml`][pyyaml] - Required for API Schema and Docs auto generation.\n* [`apispec`][apispec] - Required for API Schema and Docs auto generation.\n* [`python-forge`][python-forge] - Required for pagination.\n* [`sqlalchemy`][SQLAlchemy] - Required for Generic API resources.\n* [`databases`][databases] - Required for Generic API resources.\n\nYou can install all of these with `pip3 install flama[full]`.\n\n## Credits\n\nThat library is heavily inspired by [APIStar] server in an attempt to bring a good amount of it essence to work with \n[Starlette] as the ASGI framework and [Marshmallow] as the schema system.\n\n## Contributing\n\nThis project is absolutely open to contributions so if you have a nice idea, create an issue to let the community \ndiscuss it.\n\n[Python]: https://www.python.org\n[Starlette]: https://starlette.io\n[APIStar]: https://github.com/encode/apistar/tree/version-0.5.x\n[Marshmallow]: https://marshmallow.readthedocs.io/\n[Swagger UI]: https://swagger.io/tools/swagger-ui/\n[ReDoc]: https://rebilly.github.io/ReDoc/\n[pyyaml]: https://pyyaml.org/wiki/PyYAMLDocumentation\n[apispec]: https://apispec.readthedocs.io/\n[python-forge]: https://python-forge.readthedocs.io/\n[SQLAlchemy]: https://www.sqlalchemy.org/\n[databases]: https://github.com/encode/databases',
    'author': 'José Antonio Perdiguero López',
    'author_email': 'perdy@perdy.io',
    'url': 'https://github.com/PeRDy/flama',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from flama import exceptions, http
from flama.codecs.base import HTTPCodec

__all__ = ["JSONDataCodec"]


class JSONDataCodec(HTTPCodec):
    media_type = "application/json"
    format = "json"

    async def decode(self, request: http.Request, **options):
        try:
            if await request.body() == b"":
                return None

            return await request.json()
        except ValueError as exc:
            raise exceptions.DecodeError(f"Malformed JSON. {exc}") from None



import asyncio
import inspect
import typing
from functools import wraps

import marshmallow
import starlette.routing
from starlette.concurrency import run_in_threadpool
from starlette.routing import Match, Mount
from starlette.types import ASGIApp, Receive, Scope, Send

from flama import http, websockets
from flama.components import Component
from flama.responses import APIResponse
from flama.types import Field, FieldLocation, HTTPMethod, OptBool, OptFloat, OptInt, OptStr
from flama.validation import get_output_schema

if typing.TYPE_CHECKING:
    from flama.resources import BaseResource

__all__ = ["Route", "WebSocketRoute", "Router"]


FieldsMap = typing.Dict[str, Field]
MethodsMap = typing.Dict[str, FieldsMap]

PATH_SCHEMA_MAPPING = {
    inspect.Signature.empty: lambda *args, **kwargs: None,
    int: marshmallow.fields.Integer,
    float: marshmallow.fields.Number,
    str: marshmallow.fields.String,
    bool: marshmallow.fields.Boolean,
    http.PathParam: marshmallow.fields.String,
}

QUERY_SCHEMA_MAPPING = {
    inspect.Signature.empty: lambda *args, **kwargs: None,
    int: marshmallow.fields.Integer,
    float: marshmallow.fields.Number,
    bool: marshmallow.fields.Boolean,
    str: marshmallow.fields.String,
    OptInt: marshmallow.fields.Integer,
    OptFloat: marshmallow.fields.Number,
    OptBool: marshmallow.fields.Boolean,
    OptStr: marshmallow.fields.String,
    http.QueryParam: marshmallow.fields.String,
}


class FieldsMixin:
    def _get_fields(
        self, router: "Router"
    ) -> typing.Tuple[MethodsMap, MethodsMap, typing.Dict[str, Field], typing.Dict[str, typing.Any]]:
        query_fields: MethodsMap = {}
        path_fields: MethodsMap = {}
        body_field: typing.Dict[str, Field] = {}
        output_field: typing.Dict[str, typing.Any] = {}

        if hasattr(self, "methods") and self.methods is not None:
            if inspect.isclass(self.endpoint):  # HTTP endpoint
                methods = [(m, getattr(self.endpoint, m.lower() if m != "HEAD" else "get")) for m in self.methods]
            else:  # HTTP function
                methods = [(m, self.endpoint) for m in self.methods] if self.methods else []
        else:  # Websocket
            methods = [("GET", self.endpoint)]

        for m, h in methods:
            query_fields[m], path_fields[m], body_field[m], output_field[m] = self._get_fields_from_handler(h, router)

        return query_fields, path_fields, body_field, output_field

    def _get_parameters_from_handler(
        self, handler: typing.Callable, router: "Router"
    ) -> typing.Dict[str, inspect.Parameter]:
        parameters = {}

        for name, parameter in inspect.signature(handler).parameters.items():
            for component in router.components:
                if component.can_handle_parameter(parameter):
                    parameters.update(self._get_parameters_from_handler(component.resolve, router))
                    break
            else:
                parameters[name] = parameter

        return parameters

    def _get_fields_from_handler(
        self, handler: typing.Callable, router: "Router"
    ) -> typing.Tuple[FieldsMap, FieldsMap, Field, typing.Any]:
        query_fields: FieldsMap = {}
        path_fields: FieldsMap = {}
        body_field: Field = None

        # Iterate over all params
        for name, param in self._get_parameters_from_handler(handler, router).items():
            if name in ("self", "cls"):
                continue

            # Matches as path param
            if name in self.param_convertors.keys():
                try:
                    schema = PATH_SCHEMA_MAPPING[param.annotation]
                except KeyError:
                    schema = marshmallow.fields.String

                path_fields[name] = Field(
                    name=name, location=FieldLocation.path, schema=schema(required=True), required=True
                )

            # Matches as query param
            elif param.annotation in QUERY_SCHEMA_MAPPING:
                if param.annotation in (OptInt, OptFloat, OptBool, OptStr) or param.default is not param.empty:
                    required = False
                    kwargs = {"missing": param.default if param.default is not param.empty else None}
                else:
                    required = True
                    kwargs = {"required": True}

                query_fields[name] = Field(
                    name=name,
                    location=FieldLocation.query,
                    schema=QUERY_SCHEMA_MAPPING[param.annotation](**kwargs),
                    required=required,
                )

            # Body params
            elif inspect.isclass(param.annotation) and issubclass(param.annotation, marshmallow.Schema):
                body_field = Field(name=name, location=FieldLocation.body, schema=param.annotation())

        output_field = inspect.signature(handler).return_annotation

        return query_fields, path_fields, body_field, output_field


class Route(starlette.routing.Route, FieldsMixin):
    def __init__(self, path: str, endpoint: typing.Callable, router: "Router", *args, **kwargs):
        super().__init__(path, endpoint=endpoint, **kwargs)

        # Replace function with another wrapper that uses the injector
        if inspect.isfunction(endpoint) or inspect.ismethod(endpoint):
            self.app = self.endpoint_wrapper(endpoint)

        if self.methods is None:
            self.methods = [m for m in HTTPMethod.__members__.keys() if hasattr(self, m.lower())]

        self.query_fields, self.path_fields, self.body_field, self.output_field = self._get_fields(router)

    def endpoint_wrapper(self, endpoint: typing.Callable) -> ASGIApp:
        """
        Wraps a http function into ASGI application.
        """

        @wraps(endpoint)
        async def _app(scope: Scope, receive: Receive, send: Send) -> None:
            app = scope["app"]

            route, route_scope = app.router.get_route_from_scope(scope)

            state = {
                "scope": scope,
                "receive": receive,
                "send": send,
                "exc": None,
                "app": app,
                "path_params": route_scope["path_params"],
                "route": route,
                "request": http.Request(scope, receive),
            }

            injected_func = await app.injector.inject(endpoint, state)

            if asyncio.iscoroutinefunction(endpoint):
                response = await injected_func()
            else:
                response = await run_in_threadpool(injected_func)

            # Wrap response data with a proper response class
            if isinstance(response, (dict, list)):
                response = APIResponse(content=response, schema=get_output_schema(endpoint))
            elif isinstance(response, str):
                response = APIResponse(content=response)
            elif response is None:
                response = APIResponse(content="")

            await response(scope, receive, send)

        return _app


class WebSocketRoute(starlette.routing.WebSocketRoute, FieldsMixin):
    def __init__(self, path: str, endpoint: typing.Callable, router: "Router", *args, **kwargs):
        super().__init__(path, endpoint=endpoint, **kwargs)

        # Replace function with another wrapper that uses the injector
        if inspect.isfunction(endpoint):
            self.app = self.endpoint_wrapper(endpoint)

        self.query_fields, self.path_fields, self.body_field, self.output_field = self._get_fields(router)

    def endpoint_wrapper(self, endpoint: typing.Callable) -> ASGIApp:
        """
        Wraps websocket function into ASGI application.
        """

        @wraps(endpoint)
        async def _app(scope: Scope, receive: Receive, send: Send) -> None:
            app = scope["app"]

            route, route_scope = app.router.get_route_from_scope(scope)

            state = {
                "scope": scope,
                "receive": receive,
                "send": send,
                "exc": None,
                "app": app,
                "path_params": route_scope["path_params"],
                "route": route,
                "websocket": websockets.WebSocket(scope, receive, send),
            }

            injected_func = await app.injector.inject(endpoint, state)

            kwargs = scope.get("kwargs", {})
            await injected_func(**kwargs)

        return _app


class Router(starlette.routing.Router):
    def __init__(self, components: typing.Optional[typing.List[Component]] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if components is None:
            components = []

        self.components = components

    def add_route(
        self,
        path: str,
        endpoint: typing.Callable,
        methods: typing.List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ):
        self.routes.append(
            Route(path, endpoint=endpoint, methods=methods, name=name, include_in_schema=include_in_schema, router=self)
        )

    def add_websocket_route(self, path: str, endpoint: typing.Callable, name: str = None):
        self.routes.append(WebSocketRoute(path, endpoint=endpoint, name=name, router=self))

    def add_resource(self, path: str, resource: "BaseResource"):
        # Handle class or instance objects
        if inspect.isclass(resource):  # noqa
            resource = resource()

        for name, route in resource.routes.items():
            route_path = path + resource._meta.name + route._meta.path
            route_func = getattr(resource, name)
            name = route._meta.name if route._meta.name is not None else f"{resource._meta.name}-{route.__name__}"
            self.add_route(route_path, route_func, route._meta.methods, name, **route._meta.kwargs)

    def mount(self, path: str, app: ASGIApp, name: str = None) -> None:
        if isinstance(app, Router):
            app.components = self.components

        path = path.rstrip("/")
        route = Mount(path, app=app, name=name)
        self.routes.append(route)

    def route(
        self, path: str, methods: typing.List[str] = None, name: str = None, include_in_schema: bool = True
    ) -> typing.Callable:
        def decorator(func: typing.Callable) -> typing.Callable:
            self.add_route(path, func, methods=methods, name=name, include_in_schema=include_in_schema)
            return func

        return decorator

    def websocket_route(self, path: str, name: str = None) -> typing.Callable:
        def decorator(func: typing.Callable) -> typing.Callable:
            self.add_websocket_route(path, func, name=name)
            return func

        return decorator

    def resource(self, path: str) -> typing.Callable:
        def decorator(resource: "BaseResource") -> "BaseResource":
            self.add_resource(path, resource=resource)
            return resource

        return decorator

    def get_route_from_scope(self, scope) -> typing.Tuple[Route, typing.Optional[typing.Dict]]:
        partial = None

        for route in self.routes:
            if isinstance(route, Mount):
                path = scope.get("path", "")
                root_path = scope.pop("root_path", "")
                scope["path"] = root_path + path

            match, child_scope = route.matches(scope)
            if match == Match.FULL:
                scope.update(child_scope)

                if isinstance(route, Mount):
                    route, mount_scope = route.app.get_route_from_scope(scope)
                    return route, mount_scope

                return route, scope
            elif match == Match.PARTIAL and partial is None:
                partial = route
                partial_scope = child_scope

        if partial is not None:
            scope.update(partial_scope)
            return partial, scope

        return self.not_found, None
